""" This module is loaded instead of ufl when executing .ufl files.
So, this module contains all our extensions and monkey patches to
UFL.
"""

import ufl

from ufl import *


# Monkey patch to the FiniteElementBase
# We want to subclass finite elements in order to allow meta data
# connected to constraints. This is an intermediate solution until
# we have a proper notion of (global) function spaces in UFL. As is,
# two finite elements with different meta data would not compare equally,
# but they should in order to allow hashing generation magic based on
# them (as done everywhere)
ufl.FiniteElementBase.__eq__ = lambda s, o: repr(s) == repr(o)


class TrialFunction(ufl.Coefficient):
    """ A coefficient that always takes the reserved index 0 """
    def __init__(self, element, count=None):
        if count and count is not 0:
            raise ValueError("The trial function must be the coefficient of index 0 in uflpdelab")
        ufl.Coefficient.__init__(self, element, count=0)


class Coefficient(ufl.Coefficient):
    """ A coefficient that honors the reserved index 0. """
    def __init__(self, element, count=None):
        if count and count is 0:
            raise ValueError("The coefficient of index 0 is reserved for the trial function in uflpdelab")
        if count and count is 1:
            raise ValueError("The coefficient of index 1 is reserved for the jacobian apply vector in uflpdelab")
        if not count and ufl.Coefficient._globalcount < 2:
            count = 2
        ufl.Coefficient.__init__(self, element, count)


def split(obj):
    if isinstance(obj, Expression):
        return split_expression(obj)
    return ufl.split_functions.split2(obj)


def Coefficients(element):
    return split(Coefficient(element))


def TestFunctions(element):
    return split(TestFunction(element))


def TrialFunctions(element):
    return split(TrialFunction(element))


class Expression(Coefficient):
    def __init__(self, cppcode=None, element=None, cell="triangle", degree=None, is_global=True, on_intersection=False, cellwise_constant=False):
        assert cppcode

        if isinstance(cppcode, str):
            cppcode = (cppcode,)

        def wrap_return(e):
            if "return" not in e:
                return "return {};".format(e)
            else:
                return e

        cppcode = tuple(wrap_return(e) for e in cppcode)

        if cellwise_constant:
            if element:
                assert element.degree() == 0
            elif degree is not None:
                assert degree == 0
            else:
                element = FiniteElement("DG", cell, 0)

        if degree is None:
            degree = 1

        if element is None:
            if len(cppcode) == 1:
                element = FiniteElement("DG", cell, degree)
            else:
                element = VectorElement("DG", cell, degree, len(cppcode))

        def element_length(elem):
            if isinstance(elem, ufl.FiniteElement):
                return 1
            else:
                return elem.value_shape()[0]

        assert element_length(element) == len(cppcode)

        self.c_expr = cppcode
        self.is_global = is_global
        self.on_intersection = on_intersection
        self.element = element

        # Initialize a coefficient with a dummy finite element map.
        Coefficient.__init__(self, element)

    def __mul__(self, other):
        # Allow the combination of Expressions
        if isinstance(other, Expression):
            from ufl import MixedElement
            return Expression(cppcode=self.c_expr + other.c_expr, element=self.element * other.element)
        else:
            return Coefficient.__mul__(self, other)

    # TODO the subdomain_data code needs a uflid, not idea how to get it here
    # The standard way through class decorator fails here...
    def ufl_id(self):
        return 0


def split_expression(expr):
    assert isinstance(expr, Expression)

    def element_slice(expression, sub):
        offset = sum(subel.value_size() for subel in expression.element.sub_elements()[:sub])
        return expression.c_expr[offset:offset + expr.element.sub_elements()[sub].value_size()]

    return tuple(Expression(cppcode=element_slice(expr, i),
                            element=expr.element.sub_elements()[i])
                 for i in range(expr.element.num_sub_elements())
                 )


class FiniteElement(ufl.FiniteElement):
    def __init__(self, *args, **kwargs):
        if ('dirichlet_constraints' in kwargs) or ('dirichlet_expression' in kwargs):
            # Get dirichlet_constraints and convert it to Expression if necessary!
            self.dirichlet_constraints = kwargs.pop('dirichlet_constraints', 'return true;')
            if isinstance(self.dirichlet_constraints, str):
                self.dirichlet_constraints = Expression(self.dirichlet_constraints)
            assert isinstance(self.dirichlet_constraints, Expression)

            # Get dirichlet_constraints and convert it to Expression if necessary!
            self.dirichlet_expression = kwargs.pop('dirichlet_expression', 'return 0.0;')
            if isinstance(self.dirichlet_expression, str):
                self.dirichlet_expression = Expression(self.dirichlet_expression)
            assert isinstance(self.dirichlet_expression, Expression)

        # Initialize the original finite element from ufl
        ufl.FiniteElement.__init__(self, *args, **kwargs)


class VectorElement(ufl.VectorElement):
    def __init__(self, *args, **kwargs):
        if ('dirichlet_constraints' in kwargs) or ('dirichlet_expression' in kwargs):
            # Get dirichlet_constraints and convert it to Expression if necessary!
            self.dirichlet_constraints = kwargs.pop('dirichlet_constraints', 'return true;')
            if isinstance(self.dirichlet_constraints, str):
                self.dirichlet_constraints = Expression(self.dirichlet_constraints)
            assert isinstance(self.dirichlet_constraints, Expression)

            # Get dirichlet_constraints and convert it to Expression if necessary!
            self.dirichlet_expression = kwargs.pop('dirichlet_expression', 'return 0.0;')
            if isinstance(self.dirichlet_expression, str):
                self.dirichlet_expression = Expression(self.dirichlet_expression)
            assert isinstance(self.dirichlet_expression, Expression)

        # Initialize the original finite element from ufl
        ufl.VectorElement.__init__(self, *args, **kwargs)
